"""
Option class represents a single option in a list of options/enum
"""
import six
from .mysix import is_identifier

NUMBER_TYPES = six.integer_types + (float,)
DATE_TYPES = ()  # (datetime, date, time)

AVAILABLE_CODE_TYPES = six.string_types + NUMBER_TYPES + DATE_TYPES
AVAILABLE_CODE_TYPES_STR = ', '.join(t.__name__ for t in AVAILABLE_CODE_TYPES)


class OptionPseudo(object):

    def __init__(self, o):
        self._tags = set()
        self.tag_added = None
        self.tag_removed = None

    @property
    def tags(self):
        return tuple(self._tags)

    def add_tag(self, tag):
        self._tags.add(tag)
        if callable(self.tag_added):
            self.tag_added(tag)

    def remove_tag(self, tag):
        self._tags.remove(tag)
        if callable(self.tag_removed):
            self.tag_removed(tag)

    def get_text(self):
        return self.text if self.text is not None else self.name


class OptionMeta(type):

    def __new__(mcs, name, bases, namespace):
        cls_instance = super(OptionMeta, mcs).__new__(mcs, name, bases, namespace)

        cls_instance.NOT_DEFINED = None
        """Not defined option."""

        cls_instance.NUMBER_TYPES = six.integer_types + (float,)
        cls_instance.DATE_TYPES = ()  # (datetime, date, time)

        cls_instance.AVAILABLE_CODE_TYPES = six.string_types + NUMBER_TYPES + DATE_TYPES
        cls_instance.AVAILABLE_CODE_TYPES_STR = ', '.join(t.__name__ for t in AVAILABLE_CODE_TYPES)

        return cls_instance

    def __call__(cls, code, name, text=None, tags=None):

        if code is None or name is None:
            raise ValueError('code or name can not be None')

        if not (code is None or isinstance(code, Option.AVAILABLE_CODE_TYPES)):
            raise TypeError('Option code must be one of %s' % AVAILABLE_CODE_TYPES_STR)

        if name is not None:
            if not isinstance(name, six.string_types):
                raise ValueError('Option name must be string type.')

            if not (name[0].isalpha() and is_identifier(name) and name.isupper()):
                raise ValueError('Option name must be alphanumeric or "_"  in uppercase and start with alphabet.')

        if not ((code is None and name is None and text is None) or (code is not None and name is not None)):
            raise ValueError('Option code and name must be not None unless code, name and text are all None.')

        if tags is not None and not isinstance(tags, (tuple, list)):
            raise ValueError('"tags" must be a tuple or list of strings.')

        cls_code = type(code)
        cls_option = type('Option(%s)' % cls_code.__name__, (cls_code, OptionPseudo), {})
        obj_option = cls_option(code)
        obj_option.code = code
        obj_option.name = name
        obj_option.text = text

        if tags is not None:
            tags = set(tags)
            for tag in tags:
                if not isinstance(tag, six.string_types):
                    raise ValueError('"tags" must be a tuple or list of strings. "%s" is not a string object.' % tag)
                if not (tag.isalpha() and is_identifier(tag)):
                    raise ValueError('Tags must be alphanumeric or "_"  and start with alphabet.')
                obj_option.add_tag(tag)

        return obj_option

    def __subclasscheck__(cls, subclass):
        if issubclass(subclass, OptionPseudo):
            return True
        else:
            return super(OptionMeta, cls).__subclasscheck__(subclass)

    def __instancecheck__(cls, instance):
        if isinstance(instance, OptionPseudo):
            return True
        else:
            return super(OptionMeta, cls).__instancecheck__(instance)


@six.add_metaclass(OptionMeta)     # py2,3 compatibility
class Option(object):
    pass


__all__ = ('Option', )





"""
My py2 & py3 compilable module.
"""

import six

if six.PY2:
    import re
    regx_identifier = re.compile('^[a-z][a-z0-9_]*$', re.IGNORECASE)


def is_identifier(s):
    if six.PY2:
        m = regx_identifier.match(s)
        return bool(m)
    else:
        return s.isidentifier()



"""
Options represents an set of option items (enumerations). It is a list of options/enum entries.
"""

import six
from .option import Option
import logging

log = logging.getLogger(__name__)


class OptionGroup(list):

    def add(self, opt):
        super(OptionGroup, self).append(opt)

    def remove(self, opt):
        super(OptionGroup, self).remove(opt)


class OptionsMeta(type):

    def __new__(mcs, name, bases, namespace):
        instance = super(OptionsMeta, mcs).__new__(mcs, name, bases, namespace)
        name_options_mapping = {}
        code_options_mapping = {}

        ignore_invalid_name = namespace.get('__IGNORE_INVALID_NAME__', False)
        order_by = namespace.get('__ORDER_BY__', None)

        if not isinstance(ignore_invalid_name, bool):
            raise ValueError("'__IGNORE_INVALID_NAME__' must be bool type True or False.")

        if order_by not in ['code', 'name', None]:
            raise ValueError("'__ORDER_BY__' only supported on `code` and `name` field")

        if name != 'Options':
            for attr, val in namespace.items():
                if attr.startswith('_'):
                    continue
                elif not attr.isupper():
                    if callable(val) or isinstance(val, (staticmethod, classmethod, property)):
                        pass    # function
                    else:
                        if not ignore_invalid_name:
                            raise AttributeError('Option name must be uppercase. Attribute "%s" is not.' % attr)
                else:
                    if attr in name_options_mapping.keys() or hasattr(mcs, attr):
                        raise AttributeError('Duplicated attribute "%s" found' % attr)

                    if isinstance(val, Option):
                        if val.name != attr:
                            raise ValueError('Option name of option %s must be same as attribute "%s"' % (val, attr))
                        else:
                            opt = val
                    elif isinstance(val, (list, tuple)):
                        if len(val) == 0:
                            raise ValueError('Option code can not be empty list or tuple')
                        elif len(val) == 1:
                            opt = Option(code=val[0], name=attr)
                        elif len(val) == 2:
                            opt = Option(code=val[0], name=attr, text=str(val[1]))
                        elif len(val) == 3:
                            opt = Option(code=val[0], name=attr, text=str(val[1]), tags=val[2])
                        else:
                            raise ValueError('Option code can only be (code, text) tuple or list. name is not needed.')
                    elif isinstance(val, Option.AVAILABLE_CODE_TYPES):
                        opt = Option(code=val, name=attr)
                    else:
                        raise AttributeError('"%s" is uppercase attribute but its value can not be converted to Option.'
                                             % attr)

                    if opt.code in code_options_mapping.keys():
                        raise ValueError('Duplicated code "%s" found' % opt.code)

                    setattr(instance, attr, opt)
                    name_options_mapping[attr] = opt
                    code_options_mapping[opt.code] = opt
                    mcs.__scan_option_tags_for_group(instance, opt)

        instance.__name_options_mapping__ = name_options_mapping
        instance.__code_options_mapping__ = code_options_mapping

        return instance

    def __scan_option_tags_for_group(cls, opt):
        """
        Scan all tags in a given Option object and add the object to appropriate group of the tag.
        :param opt: Option object
        :return:
        """

        def add_option_to_group(tag):
            __tag = '__%s' % tag
            group = getattr(cls, __tag, None)
            if group is not None:
                if isinstance(group, OptionGroup):
                    group.add(opt)
                else:
                    raise ValueError('Tag "%s" is duplicated as attribute of "%s"'
                                     % (tag, cls.__name__))
            else:
                group = OptionGroup()
                group.add(opt)
                setattr(cls, __tag, group)
            setattr(cls, tag, tuple(group))

        def remove_option_from_group(tag):
            __tag = '__%s' % tag
            group = getattr(cls, __tag, None)
            if group is not None:
                if isinstance(group, OptionGroup):
                    group.remove(opt)
                else:
                    raise ValueError('No options are grouped in with "%s" in "%s"' % (tag, cls.__name__))
            else:
                raise ValueError('Option group for tag "%s" is not existing in "%s"' % (tag, cls.__name__))
            setattr(cls, tag, tuple(group))

        for tag in opt.tags:
            add_option_to_group(tag)

        opt.tag_added = add_option_to_group
        opt.tag_removed = remove_option_from_group

    # Options class methods or properties
    def __get_name_options_mapping(cls):
        """
        Dict of {'name': option} mapping.
        DO NOT change this dict. TODO: make it immutable.

        :return: dict {"name": option}
        """
        return cls.__name_options_mapping__

    def __get_code_options_mapping(cls):
        """
        Dict of {'code': option} mapping
        DO NOT change this dict. TODO: make it immutable.
        :return: dict {"name": option}
        """
        return cls.__code_options_mapping__

    def __getitem__(cls, item):
        return cls.__get_name_options_mapping()[item]

    def __setitem__(cls, key, value):
        raise TypeError("'%s' class does not support item assignment" % cls.__name__)

    def __delitem__(cls, key):
        raise TypeError("'%s' class does not support item deletion" % cls.__name__)

    def __contains__(cls, item):
        if isinstance(item, Option):
            return item in cls.__get_name_options_mapping().values()
        else:
            return item in cls.__get_code_options_mapping().keys()

    def get(cls, key, default=None):
        return cls.__get_name_options_mapping().get(key, default)

    @property
    def codes(cls):
        """List of `code`s"""
        return list(cls.__get_code_options_mapping().keys())

    @property
    def names(cls):
        """List of `name`s"""
        return list(cls.__get_name_options_mapping().keys())

    @property
    def all(cls):
        """List of `Option` objects"""
        return list(cls.__get_name_options_mapping().values())

    @property
    def tuples(cls):
        """List of (`name`, `code`, `text`) tuples"""
        return [(o.name, o.code, o.text) for o in cls.__get_name_options_mapping().values()]

    @property
    def items(cls):
        """Dict of {`name`: Option} mapping"""
        return {o.name: o for o in cls.__get_name_options_mapping().values()}

    def get_list(cls, *fields):
        """
        List of *fields.
        :param fields: Field name of `Option`. Such as `code`, (`name`, `code`), [`code`, `name`, `text`] or etc.
        :return: List of *fields tuple or single value.
        """

        if len(fields) == 0:
            raise ValueError("No fields argument found.")

        found_fields = set()
        for f in fields:
            if f not in ['code', 'name', 'text']:
                raise NameError(
                    "'%s' is incorrect Option field. Only 'code', 'name' and 'text' are available." % str(f))
            if f in found_fields:
                raise ValueError("Duplicated fields '%s' found." % str(f))
            found_fields.add(f)

        lst = []
        for o in cls.__get_name_options_mapping().values():
            value = []
            for f in fields:
                value.append(getattr(o, f))
            lst.append(value[0] if len(value) == 1 else tuple(value))
        return lst

    def get_dict(cls, key_field, *fields):
        """
        Retrieve as dict of {key_field: *fields} mapping. If `fields` is single value, returns `{key:value}` mapping.
                If `fields` is **tuple**, returns `{key: (tuple of fields value)}` mapping.
        :param key_field: name of Option field for dict key. Only `code` and `name` are available.
        :param fields: Names of Option field for values. Can be tuple or single value. Impact value part of return dict.
        :return: dict of {key_field: *fields} mapping.
        """
        if key_field not in ['code', 'name']:
            raise NameError("'%s' is not correct key field. Only 'code' and 'name' can be key field." % str(key_field))

        if len(fields) == 0:
            raise ValueError("No fields argument found.")

        found_fields = set()
        for f in fields:
            if f not in ['code', 'name', 'text']:
                raise NameError(
                    "'%s' is incorrect Option field. Only 'code', 'name' and 'text' are available." % str(f))
            if f in found_fields:
                raise ValueError("Duplicated fields '%s' found." % str(f))
            found_fields.add(f)

        items = {}
        for o in cls.__get_name_options_mapping().values():
            key = getattr(o, key_field)
            value = []
            for f in fields:
                value.append(getattr(o, f))
            items[key] = value[0] if len(value) == 1 else tuple(value)

        return items


@six.add_metaclass(OptionsMeta)     # py2,3 compatibility
class Options(object):
    """
    Derive class `Options` to implement your enum/options.
    e.g
    ```
        # simple way
        class Fruit(Options):
            APPLE = 1
            ORANGE = 2
            BANANA = 3

        # option `code` + `text` description in tuple
        class DoorState(Options):
            OPEN = 'O', 'Door is opened'
            CLOSED = ('C', 'Door is closed')
            IN_OPENING = 'IO'
            IN_CLOSING = 'IC'
    ```
    """

    pass


__all__ = ('Options', )




"""
optenum
==========

A missing python Enum/option lib supports enum code, name, text, even (code, name) tuple list and so on.

Compatible with Python 2.7+ and Python 3.0+.

Features:

  * Code - Enumration/option by different types - e.g. 0, 1, -1 (or 'new', 'running', 'stopped')
  * Name - Name of an enum/option - e.g.  'NEW', 'RUNNING', 'STOPPED'. Support dot access.
  * Text - Meaning or description of the enum/option. support i18n - e.g. _('new'), _('running'), _('stopped')
  * List - Retrieve list of code, name or text `[0, 1, -1]`
  * Dict - Retrieve dict of `{code: name}` mapping. even `{code: text}`, `{name: text}` mapping if required.
  * List of tuples - Retrieve list of `[(code, name), ...]` tuples. Useful in **Django** model.
  * Grouping - Group a set of enums/options. e.g. IN_PROGERSS_STATE = ['STARTING', 'STOPPING'], but 'STARTED' and 'STOPPED' are not belongs to it.
  * Access **name**, **text** by **code**
  * Lookup enum/option by **name**, **code**

"""

from __future__ import absolute_import
from .version import __version__
from .option import Option
from .options import Options

__all__ = ('Option', 'Options', '__version__')

__copyright__ = "Copyright (c) 2019 Samuel Chen (Chen Wei)"
__license__ = "MIT"
__summary__ = "A missing python Enum/option lib supports enum code, name, text, even (code, name) tuple list and so on."
__uri__ = "https://github.com/samuelchen/optenum.git"



"""
Option class represents a single option in a list of options/enum

DEPRECATED from v1.1.0 version.
"""
import six
from datetime import datetime, date, time
from .mysix import is_identifier


class Option(object):
    """ A single option with code, name and text"""

    NOT_DEFINED = None
    """Not defined option."""

    NUMBER_TYPES = six.integer_types + (float, )
    DATE_TYPES = ()    # (datetime, date, time)

    AVAILABLE_CODE_TYPES = six.string_types + NUMBER_TYPES + DATE_TYPES
    AVAILABLE_CODE_TYPES_STR = ', '.join(t.__name__ for t in AVAILABLE_CODE_TYPES)

    def __init__(self, code, name, text=None):
        """
        Initialize an option
        :param code: The real value of an option or enum entry.
        :param name: The readable name of an option or enum entry. It can also be used as dot annotation.
                        It is alphanumeric or "_"  in uppercase and start with alphabet.
        :param text: The description of the option/enum entry. It's used to display. Can be i18n.
        """

        if not (code is None or isinstance(code, self.AVAILABLE_CODE_TYPES)):
            raise TypeError('Option code must be one of %s' % self.AVAILABLE_CODE_TYPES_STR)

        if name is not None:
            if not isinstance(name, six.string_types):
                raise ValueError('Option name must be string type.')

            if not (name[0].isalpha() and is_identifier(name) and name.isupper()):
                raise ValueError('Option name must be alphanumeric or "_"  in uppercase and start with alphabet.')

        if not ((code is None and name is None and text is None) or (code is not None and name is not None)):
            raise ValueError('Option code and name must be not None unless code, name and text are all None.')

        self.__code = code
        self.__name = name
        self.__text = text

        self.__code_type = type(code)

    @property
    def code(self):
        """The real value of an option or enum entry."""
        return self.__code

    @property
    def name(self):
        """
        The readable string of an option or enum entry. It's commonly uppercase.
        It can also be used as dot annotation.
        """
        return self.__name

    @property
    def text(self):
        """The description of the option/enum entry. It's used to display. Can be i18n."""
        return self.__text

    def get_text(self):
        """
        Returns `text` of the option. If `text` is None, returns `name` in lower case as text.
        If name is also None, return `None` object converted string.
        """
        return str(None) if self.name is None else self.name.lower() if self.text is None else self.text

    def __repr__(self):
        return "<%s code=%s name=%s text=%s>" % (self.__class__.__name__, self.code, self.name, self.text)

    def __str__(self):
        return str(self.code)

    # Compare Operators

    def __op_cmp_check__(self, op, other):
        """
        Check if self is compatible with `other` on operation `op`
        :param op: operation such as "+", "-" and so on
        :param other: The other value of the operation to be compared with self.
        :return: void.  (raise error directly if not compatible)
        """
        other_is_option = False
        if isinstance(other, Option):
            other = other.code
            other_is_option = True
        type_other = type(other)

        if isinstance(other, self.NUMBER_TYPES + self.DATE_TYPES) and isinstance(self.code, self.NUMBER_TYPES + self.DATE_TYPES):
            pass
        elif isinstance(other, six.string_types) and isinstance(self.code, six.string_types):
            pass
        elif other is None or self.code is None or self == Option.NOT_DEFINED:
            pass
        else:
            raise TypeError("'%s' not supported between instances of 'Option(%s)' and '%s'" % (
                op, self.__code_type.__name__,
                'Option(%s)' % type_other.__name__ if other_is_option else type_other.__name__
            ))

    # if six.PY2:
    #     def __cmp__(self, other):
    #         self.__op_cmp_check__('==', other)
    #
    #         if isinstance(other, Option):
    #             if self.code == other.code and self.name == other.name and self.text == other.text:
    #                 return 0
    #             elif self.code < other.code:
    #                 return -1
    #             else:
    #                 return 1
    #         else:
    #             return self.code.__cmp__(other)

    def __eq__(self, other):
        self.__op_cmp_check__('==', other)

        if isinstance(other, Option):
            return self.code == other.code and self.name == other.name and self.text == other.text
        else:
            return self.code == other

    def __hash__(self):
        """return hash of `code` to ensure key access to `set`, `dict` or other hash based collection"""
        # return super(Option, self).__hash__()
        return self.code.__hash__()

    def __lt__(self, other):
        self.__op_cmp_check__('<', other)

        return self.code < (other.code if isinstance(other, Option) else other)

    def __le__(self, other):
        self.__op_cmp_check__('<=', other)

        return self.code <= (other.code if isinstance(other, Option) else other)

    def __gt__(self, other):
        self.__op_cmp_check__('>', other)

        return self.code > (other.code if isinstance(other, Option) else other)

    def __ge__(self, other):
        self.__op_cmp_check__('>=', other)

        return self.code >= (other.code if isinstance(other, Option) else other)

    # Math Operators

    def __op_math_check__(self, op, other):
        """
        Check if self is compatible with `other` on math operation `op`
        :param op: operation such as "+", "-" and so on
        :param other: The other value of the operation to be compared with self.
        :return: void.  (raise error directly if not compatible)
        """
        other_is_option = False
        if isinstance(other, Option):
            other = other.code
            other_is_option = True
        type_other = type(other)

        if isinstance(other, self.NUMBER_TYPES) and isinstance(self.code, self.NUMBER_TYPES):
            pass
        elif isinstance(other, six.string_types) and isinstance(self.code, six.string_types):
            pass
        elif isinstance(other, self.DATE_TYPES) and isinstance(self.code, self.DATE_TYPES):
            pass
        else:
            raise TypeError("'%s' not supported between instances of 'Option(%s)' and '%s'" % (
                op, self.__code_type.__name__,
                'Option(%s)' % type_other.__name__ if other_is_option else type_other.__name__
            ))

    def __op_num_check__(self, op, val):

        val, val_is_option = (val.code, True) if isinstance(val, Option) else (val, False)
        type_val = type(val)
        type_str = 'Option(%s)' % type_val.__name__ if val_is_option else type_val.__name__

        if not isinstance(val, self.NUMBER_TYPES):
            raise TypeError("'%s' not supported on instances of '%s'" % (op, type_str))

    def __op_int_check__(self, op, val):

        val, val_is_option = (val.code, True) if isinstance(val, Option) else (val, False)
        type_val = type(val)
        type_str = 'Option(%s)' % type_val.__name__ if val_is_option else type_val.__name__

        if not isinstance(val, six.integer_types):
            raise TypeError("'%s' not supported on instances of '%s'" % (op, type_str))

    def __neg__(self):
        self.__op_num_check__('-(negative)', self)

        return - self.code

    def __pos__(self):
        self.__op_num_check__('+(positive)', self)

        return + self.code

    def __abs__(self):
        self.__op_num_check__('abs', self)

        return abs(self.code)

    def __int__(self):
        return int(self.code)

    def __float__(self):
        return float(self.code)

    def __invert__(self):
        self.__op_int_check__('~', self)

        return ~ self.code

    def __lshift__(self, other):
        self.__op_int_check__('<<', self)
        self.__op_int_check__('<<', other)

        return self.code << other

    def __rlshift__(self, other):
        self.__op_int_check__('<<', self)
        self.__op_int_check__('<<', other)

        return other << self.code

    def __rshift__(self, other):
        self.__op_int_check__('>>', self)
        self.__op_int_check__('>>', other)

        return self.code >> other

    def __rrshift__(self, other):
        self.__op_int_check__('>>', self)
        self.__op_int_check__('>>', other)

        return other >> self.code

    def __add__(self, other):
        self.__op_math_check__('+', other)

        return self.code + (other.code if isinstance(other, Option) else other)

    def __radd__(self, other):
        return self.__add__(other)

    def __sub__(self, other):
        self.__op_math_check__('-', other)

        return self.code - (other.code if isinstance(other, Option) else other)

    def __rsub__(self, other):
        self.__op_math_check__('-', other)

        return (other.code if isinstance(other, Option) else other) - self.code

    def __mul__(self, other):
        self.__op_num_check__('*', other)

        return self.code * (other.code if isinstance(other, Option) else other)

    def __rmul__(self, other):
        return self.__mul__(other)

    def __truediv__(self, other):
        self.__op_num_check__('/', self)
        self.__op_num_check__('/', other)

        return self.code.__truediv__(other.code if isinstance(other, Option) else other)

    def __rtruediv__(self, other):
        self.__op_num_check__('/', self)
        self.__op_num_check__('/', other)

        return self.code.__rtruediv__(other.code if isinstance(other, Option) else other)

    def __floordiv__(self, other):
        self.__op_num_check__('//', self)
        self.__op_num_check__('//', other)

        return self.code.__floordiv__(other.code if isinstance(other, Option) else other)

    def __rfloordiv__(self, other):
        self.__op_num_check__('//', self)
        self.__op_num_check__('//', other)

        return self.code.__rfloordiv__(other.code if isinstance(other, Option) else other)

    # compatible with PY2 without "from __future__ import division"
    def __div__(self, other):
        self.__op_num_check__('/', self)
        self.__op_num_check__('/', other)

        return self.code.__div__(other.code if isinstance(other, Option) else other)

    def __rdiv__(self, other):
        self.__op_num_check__('/', self)
        self.__op_num_check__('/', other)

        return self.code.__rdiv__(other.code if isinstance(other, Option) else other)

    def __divmod__(self, other):
        self.__op_num_check__('%', self)
        self.__op_num_check__('%', other)

        return self.code.__divmod__(other.code if isinstance(other, Option) else other)

    def __rdivmod__(self, other):
        self.__op_num_check__('%', self)
        self.__op_num_check__('%', other)

        return self.code.__rdivmod__(other.code if isinstance(other, Option) else other)

    def __mod__(self, other):

        self.__op_num_check__('%', self)
        self.__op_num_check__('%', other)

        return self.code.__mod__(other.code if isinstance(other, Option) else other)

    def __rmod__(self, other):
        self.__op_num_check__('%', self)
        self.__op_num_check__('%', other)

        return self.code.__rmod__(other.code if isinstance(other, Option) else other)


Option.NOT_DEFINED = Option(None, None, None)


__all__ = ('Option', )
